// parameter access functions
#include <cpar.h>
#include <cpar_internal.h>

// ==============================================================
// Return the gpointer for the value corresponding to a key
// Returns NULL if the key isn't found
// ==============================================================
gpointer _cpar_get_raw(char *file, uint line, cpar_table *vp, char *key)
{
cpar_parameter_struct *parameter;

if (vp == NULL) _cpar_abort(file, line, "parameter table is NULL (key = \"%s\")", key);

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

parameter = (cpar_parameter_struct *)g_hash_table_lookup((*vp)->hash_table, key);
parameter->is_seen = true;
if (parameter == NULL) return NULL;
else return (gpointer)(&(parameter->value.int_value));
}

// ==============================================================
// Get the value corresponding to a key in the correct C type
// Returns 0 if the key isn't found, 1 if it is found
// Abort with error if wrong type
// ==============================================================
#define CPAR_ASSIGN_PARAMETER_TEMPLATE(ctype, ctype_name) \
bool _cpar_assign_parameter_ ## ctype_name(char *file, uint line, cpar_table *vp, char *key, ctype *value, uint type) \
{ \
cpar_parameter_struct *parameter; \
if (vp == NULL) return 0; \
_cpar_assert_cpar_table(file, line, vp); \
parameter = (cpar_parameter_struct *)g_hash_table_lookup((*vp)->hash_table, key); \
if (parameter != NULL) \
  { \
    if (CPAR_TYPE_MISMATCH_IS_FATAL && (parameter->type != type)) \
      { \
        _cpar_abort(file, line, \
              "key \"%s\": types don't match: parameter originally type %d:%s, cpar_assign_parameter was asked for type %d:%s", \
              key, parameter->type, _cpar_type_name(parameter->type), type, _cpar_type_name(type)); \
      } \
    *value = *((ctype *)(&(parameter->value.int_value))); \
    parameter->is_seen = true; \
    return true; \
  } \
else \
  { \
    return false; \
  } \
}
CPAR_ASSIGN_PARAMETER_TEMPLATE(bool, bool)
CPAR_ASSIGN_PARAMETER_TEMPLATE(int, int)
CPAR_ASSIGN_PARAMETER_TEMPLATE(uint, uint)
CPAR_ASSIGN_PARAMETER_TEMPLATE(double, double)
CPAR_ASSIGN_PARAMETER_TEMPLATE(real, real)
CPAR_ASSIGN_PARAMETER_TEMPLATE(bool*, bool_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(char*, string)
CPAR_ASSIGN_PARAMETER_TEMPLATE(char*, char_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(int*, int_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(uint*, uint_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(double*, double_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(real*, real_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(cpar_table*, cpar_pointer)
CPAR_ASSIGN_PARAMETER_TEMPLATE(gpointer, gpointer)


// ==============================================================
// Return the value corresponding to a key in the correct C type
// Fail with error message if the key isn't found
// Fail with error message if the type is incorrect
// ==============================================================
#define CPAR_GET_PARAMETER_TEMPLATE(ctype, ctype_name) \
ctype _cpar_get_parameter_ ## ctype_name(char *file, uint line, cpar_table *vp, char *key, uint type) \
{ \
cpar_parameter_struct *parameter; \
if (vp == NULL) _cpar_abort(file, line, "parameter table is NULL (key = \"%s\")", key); \
_cpar_assert_cpar_table(file, line, vp); \
parameter = (cpar_parameter_struct *)g_hash_table_lookup((*vp)->hash_table, key); \
if (parameter != NULL) \
  { \
    if (CPAR_TYPE_MISMATCH_IS_FATAL && (parameter->type != type)) \
      { \
        _cpar_abort(file, line, \
              "key \"%s\": types don't match: parameter originally type %d:%s, cpar_get_parameter was asked for type %d:%s", \
              key, parameter->type, _cpar_type_name(parameter->type), type, _cpar_type_name(type)); \
      } \
    parameter->is_seen = true; \
    return *((ctype *)(&(parameter->value.int_value))); \
  } \
else \
  { \
    _cpar_abort(file, line, "key \"%s\" was not found", key); \
  } \
return 0; /* this is here to avoid a compiler warning only */ \
}
CPAR_GET_PARAMETER_TEMPLATE(bool, bool)
CPAR_GET_PARAMETER_TEMPLATE(int, int)
CPAR_GET_PARAMETER_TEMPLATE(uint, uint)
CPAR_GET_PARAMETER_TEMPLATE(double, double)
CPAR_GET_PARAMETER_TEMPLATE(real, real)
CPAR_GET_PARAMETER_TEMPLATE(bool*, bool_pointer)
CPAR_GET_PARAMETER_TEMPLATE(char*, string)
CPAR_GET_PARAMETER_TEMPLATE(char*, char_pointer)
CPAR_GET_PARAMETER_TEMPLATE(int*, int_pointer)
CPAR_GET_PARAMETER_TEMPLATE(uint*, uint_pointer)
CPAR_GET_PARAMETER_TEMPLATE(double*, double_pointer)
CPAR_GET_PARAMETER_TEMPLATE(real*, real_pointer)
CPAR_GET_PARAMETER_TEMPLATE(cpar_table*, cpar_pointer)
CPAR_GET_PARAMETER_TEMPLATE(gpointer, gpointer)



// ==============================================================
// Return the type of the corresponding key
// Aborts if the key isn't found
// ==============================================================
uint _cpar_get_type(char *file, uint line, cpar_table *vp, char *key)
{
cpar_parameter_struct *parameter;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

parameter = (cpar_parameter_struct *)g_hash_table_lookup((*vp)->hash_table, key);
if (parameter == NULL)
  {
    _cpar_abort(file, line, "key \"%s\" not found", key);
  }

return parameter->type;
}

