// Macros and functions that are completely internal.

#define CPAR_TYPE_MISMATCH_IS_FATAL true
#define CPAR_ERROR_STREAM stderr

// Just some unique key to give some assurance that a structure
// is really a cpar_table structure.
#define _CPAR_MAGIC_KEY         12321
#define _CPAR_MAGIC_KEY_DELETED 12322

// This struct is used to pass a lot of values through a g_hash_table_foreach*()
// call, since those calls only allow one pointer to be passed along to the target
// function
struct ghfunc_passer {
  char *file;
  uint line;
  // flag test
  bool all_set;
  // misc
  uint action;
  uint flag;
  bool verbose;
  cpar_table *vp;
  // merge stuff
  uint mode;
  cpar_table *to_vp;
  char *to_key;  /* set this to change value of key during merge, NULL to copy key */
  // compare stuff
  char *compare_string;
};
typedef struct ghfunc_passer ghfunc_passer_struct;

/*
** Note on storing values of several type in value_union:
** Several places in this code, there are some tricky pointer
** operations to be able to return a void pointer to a value
** of several different types.  An example is:
** *((double *)value
** where value is a gpointer that points to a double.
** An extra tricky one is:
** (gpointer)(*(double **)value)
** where value is a gpointer that points to a gpointer.
** Since gpointers can't be dereferenced (that's just part of
** the C standard), value is first cast to double**, dereferenced
** to double*, and then cast back to gpointer.
**
** I am making the assumtions:
** - That all pointers are the same internally
** - That the pointer to one value in a union is the same as any other
** As far as I know, all C compilers have these properties, but
** if they don't, this code will not work.
*/


