#include <cpar.h>
#include <cpar_internal.h>

// ==============================================================
// the GHFunc for functions below
// ==============================================================
void _cpar_flag_GHFunc(gpointer key, gpointer value, gpointer user_data)
{
ghfunc_passer_struct *ghfunc_passer = (ghfunc_passer_struct *)user_data;
cpar_parameter_struct *parameter = (cpar_parameter_struct *)value;

// Skip if compare exists and it doesn't match the key
if (ghfunc_passer->compare_string != NULL)
  {
    guint compare_length = strlen(ghfunc_passer->compare_string);
    if (strncmp(ghfunc_passer->compare_string, (char *)key, compare_length) != 0)
      {
        return;
      }
  }

switch(ghfunc_passer->flag)
  {
    case CPAR_FLAG_SEEN:
        switch(ghfunc_passer->action)
          {
            case CPAR_FLAG_ASSERT:
              if (parameter->is_seen == false)
                {
                  ghfunc_passer->all_set = false;
                  // Uncomment the next line if you want to abort immediately on 
                  // finding an unseen key
                  //_cpar_abort(ghfunc_passer->file,
                  //            ghfunc_passer->line, "key \"%s\" was not seen", (char *)key);
                  if (ghfunc_passer->verbose == true)
                    {
                      fprintf(CPAR_ERROR_STREAM, "%s:%d: key \"%s\" was not seen\n", 
                              ghfunc_passer->file, ghfunc_passer->line, (char *)key);
                    }
                }
              break;
            case CPAR_FLAG_SET:
              parameter->is_seen = true;
              break;
            case CPAR_FLAG_CLEAR:
              parameter->is_seen = false;
              break;
            default:
              _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, "unknown action (%d)", 
                          ghfunc_passer->action);
          } // end switch (action)
    break;
    
    case CPAR_FLAG_DEFAULT:
        switch(ghfunc_passer->action)
          {
            case CPAR_FLAG_ASSERT:
              if (parameter->is_default == false)
                {
                  ghfunc_passer->all_set = false;
                  if (ghfunc_passer->verbose == true)
                    {
                      fprintf(CPAR_ERROR_STREAM, "%s:%d: key \"%s\" is not a default\n", 
                              ghfunc_passer->file, ghfunc_passer->line, (char *)key);
                    }
                }
              break;
            case CPAR_FLAG_SET:
              parameter->is_default = true;
              break;
            case CPAR_FLAG_CLEAR:
              parameter->is_default = false;
              break;
            default:
              _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, "unknown action (%d)", 
                          ghfunc_passer->action);
          } // end switch (action)
    break;
    
    default:
      _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, 
                  "unknown flag (%d)", ghfunc_passer->flag);    
  } // end switch (flag)
} // end _cpar_is_flag_set_GHFunc()

// ==============================================================
// Perform a flag action on all keys in the list
// ==============================================================
void _cpar_flag_multi(cpar_table *vp, ghfunc_passer_struct *ghfunc_passer, va_list va)
{
char *key;
cpar_parameter_struct *parameter = NULL;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(ghfunc_passer->file, ghfunc_passer->line, vp);

key = va_arg(va, char *);
while (key != NULL)
  {
    // Get parameter for this key
    parameter = (cpar_parameter_struct *)g_hash_table_lookup((*vp)->hash_table, key);
    if (parameter == NULL) _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, 
                                       "key \"%s\" doesn't exist", key);
    _cpar_flag_GHFunc(key, parameter, ghfunc_passer);
    key = va_arg(va, char *);
  }
} // end _cpar_flag_multi()


// ==============================================================
// Return 1 if all the parameter in a parameter table have this flag set
// Print messages if verbose != 0
// ==============================================================
bool _cpar_is_all_flag_set(char *file, uint line, cpar_table *vp, uint flag, bool verbose)
{
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.all_set = true;
ghfunc_passer.action = CPAR_FLAG_ASSERT;
ghfunc_passer.flag = flag;
ghfunc_passer.verbose = verbose;
ghfunc_passer.compare_string = NULL;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

g_hash_table_foreach((*vp)->hash_table, _cpar_flag_GHFunc, (gpointer)(&ghfunc_passer));
return ghfunc_passer.all_set;
} // end _cpar_is_all_flag_set()

// ==============================================================
// Return 1 if all the parameters given have this flag set
// Print messages if verbose != 0
// ==============================================================
bool _cpar_is_flag_set(char *file, uint line, cpar_table *vp, uint flag, bool verbose, ...)
{
va_list va;
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.all_set = true;
ghfunc_passer.action = CPAR_FLAG_ASSERT;
ghfunc_passer.flag = flag;
ghfunc_passer.verbose = verbose;
ghfunc_passer.compare_string = NULL;

va_start(va, verbose);
_cpar_flag_multi(vp, &ghfunc_passer, va);
va_end(va);

return ghfunc_passer.all_set;
}

// ==============================================================
// Return 1 if all the parameters with keys matching compare_string have this flag set
// Print messages if verbose != 0
// ==============================================================
bool _cpar_is_flag_set_match(char *file, uint line, cpar_table *vp, uint flag,
                             bool verbose, char *compare_string)
{
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.all_set = true;
ghfunc_passer.action = CPAR_FLAG_ASSERT;
ghfunc_passer.flag = flag;
ghfunc_passer.verbose = verbose;
ghfunc_passer.compare_string = compare_string;

g_hash_table_foreach((*vp)->hash_table, _cpar_flag_GHFunc, (gpointer)(&ghfunc_passer));
return ghfunc_passer.all_set;
}

// ==============================================================
// set this flag for all members of this parameter table
// ==============================================================
void _cpar_set_all_flag(char *file, uint line, cpar_table *vp, uint flag, uint action)
{
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.action = action;
ghfunc_passer.flag = flag;
ghfunc_passer.verbose = false;
ghfunc_passer.compare_string = NULL;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

g_hash_table_foreach((*vp)->hash_table, _cpar_flag_GHFunc, (gpointer)(&ghfunc_passer));
}

// ==============================================================
// set this flag for all keys given
// ==============================================================
void _cpar_set_flag(char *file, uint line, cpar_table *vp, uint flag, uint action, ...)
{
va_list va;
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.action = action;
ghfunc_passer.flag = flag;
ghfunc_passer.verbose = false;
ghfunc_passer.compare_string = NULL;

va_start(va, action);
_cpar_flag_multi(vp, &ghfunc_passer, va);
va_end(va);
}

// ==============================================================
// set this flag for keys starting with string compare_string
// ==============================================================
void _cpar_set_flag_match(char *file, uint line, cpar_table *vp, 
                          uint flag, uint action, char *compare_string)
{
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.action = action;
ghfunc_passer.flag = flag;
ghfunc_passer.verbose = false;
ghfunc_passer.compare_string = compare_string;

g_hash_table_foreach((*vp)->hash_table, _cpar_flag_GHFunc, (gpointer)(&ghfunc_passer));
}


